@@ -155,6 +155,44 @@ static void update_counters(struct osmo_fsm_inst *fi, bool conn_accepted)
 	}
 }
 
+static void lu_delay_timer_cb(void *data)
+{
+	struct msc_a *msc_a = (struct msc_a *)data;
+	msc_a_put(msc_a, MSC_A_USE_LOCATION_UPDATING);
+}
+
+static void msc_a_put_lu_deferred(struct msc_a *msc_a)
+{
+	unsigned long Tval;
+
+	/* The idea behind timer X36 is to allow re-using the radio channel that was used for
+	 * Location Updating to deliver MT SMS over GSUP.  This is achieved by delaying
+	 * release of a BSSAP/RANAP connection and thus delaying the release of the radio
+	 * channel.  The delay can be configured separately for GERAN and UTRAN. */
+	switch (msc_a->c.ran->type) {
+	case OSMO_RAT_GERAN_A:
+		Tval = osmo_tdef_get(msc_tdefs_geran, -36, OSMO_TDEF_MS, 0);
+		break;
+	case OSMO_RAT_UTRAN_IU:
+		Tval = osmo_tdef_get(msc_tdefs_utran, -36, OSMO_TDEF_MS, 0);
+		break;
+	default:
+		Tval = 0;
+		break;
+	}
+
+	if (Tval == 0) {
+		/* no delay, put LU token immediately */
+		msc_a_put(msc_a, MSC_A_USE_LOCATION_UPDATING);
+		return;
+	}
+
+	LOG_MSC_A(msc_a, LOGL_INFO, "Keeping LU token for +%lu ms\n", Tval);
+	osmo_timer_schedule(&msc_a->lu_delay_timer,
+			    Tval / 1000, /* seconds */
+			    Tval % 1000 * 1000); /* microseconds */
+}
+
 static void evaluate_acceptance_outcome(struct osmo_fsm_inst *fi, bool conn_accepted)
 {
 	struct msc_a *msc_a = fi->priv;
@@ -180,9 +218,9 @@ static void evaluate_acceptance_outcome(struct osmo_fsm_inst *fi, bool conn_acce
 		osmo_signal_dispatch(SS_SUBSCR, S_SUBSCR_ATTACHED, msc_a_vsub(msc_a));
 
 	if (msc_a->complete_layer3_type == COMPLETE_LAYER3_LU)
-		msc_a_put(msc_a, MSC_A_USE_LOCATION_UPDATING);
+		msc_a_put_lu_deferred(msc_a);
 
-	if (msc_a->complete_layer3_type == COMPLETE_LAYER3_CM_RE_ESTABLISH_REQ) {
+	if (conn_accepted && msc_a->complete_layer3_type == COMPLETE_LAYER3_CM_RE_ESTABLISH_REQ) {
 		/* Trigger new Assignment to recommence the voice call. A little dance here because normally we verify
 		 * that no CC trans is already active. */
 		struct gsm_trans *cc_trans = msc_a->cc.active_trans;
@@ -635,13 +673,28 @@ int msc_a_ensure_cn_local_rtp(struct msc_a *msc_a, struct gsm_trans *cc_trans)
 				  rtp_to_ran->codecs_known ? &rtp_to_ran->codecs : NULL, NULL);
 }
 
+static void assignment_request_timeout_cb(void *data);
+
 /* The MGW has given us a local IP address for the RAN side. Ready to start the Assignment of a voice channel. */
-static void msc_a_call_leg_ran_local_addr_available(struct msc_a *msc_a)
+void msc_a_tx_assignment_cmd(struct msc_a *msc_a)
 {
 	struct ran_msg msg;
 	struct gsm_trans *cc_trans = msc_a->cc.active_trans;
 	struct gsm0808_channel_type channel_type;
 
+	/* Do not dispatch another Assignment Command before an earlier assignment is completed. This is a sanity
+	 * safeguard, ideally callers should not even invoke this function when an Assignment is already ongoing.
+	 * (There is no osmo_fsm for Assignment / the CC trans code; when we refactor that one day, this timer should be
+	 * an FSM state.) */
+	if (osmo_timer_pending(&msc_a->cc.assignment_request_pending)) {
+		LOG_MSC_A(msc_a, LOGL_ERROR,
+			  "Not transmitting Assignment, still waiting for the response to an earlier Assignment\n");
+		return;
+	}
+	osmo_timer_setup(&msc_a->cc.assignment_request_pending, assignment_request_timeout_cb, msc_a);
+	osmo_timer_schedule(&msc_a->cc.assignment_request_pending,
+			    osmo_tdef_get(msc_a->c.ran->tdefs, -37, OSMO_TDEF_S, 10), 0);
+
 	if (!cc_trans) {
 		LOG_MSC_A(msc_a, LOGL_ERROR, "No CC transaction active\n");
 		call_leg_release(msc_a->cc.call_leg);
@@ -669,6 +722,8 @@ static void msc_a_call_leg_ran_local_addr_available(struct msc_a *msc_a)
 			return;
 		}
 		break;
+	case GSM48_BCAP_ITCAP_3k1_AUDIO:
+	case GSM48_BCAP_ITCAP_FAX_G3:
 	case GSM48_BCAP_ITCAP_UNR_DIG_INF:
 		if (!cc_trans->cc.local.bearer_services.count) {
 			LOG_TRANS(cc_trans, LOGL_ERROR, "Assignment not possible, no matching bearer service: %s\n",
@@ -802,7 +857,7 @@ static void msc_a_fsm_communicating(struct osmo_fsm_inst *fi, uint32_t event, vo
 			  rtps->use_osmux ? "yes" : "no", rtps->local_osmux_cid);
 		switch (rtps->dir) {
 		case RTP_TO_RAN:
-			msc_a_call_leg_ran_local_addr_available(msc_a);
+			msc_a_tx_assignment_cmd(msc_a);
 			return;
 		case RTP_TO_CN:
 			cc_on_cn_local_rtp_port_known(rtps->for_trans);
@@ -903,6 +958,9 @@ static void msc_a_fsm_releasing_onenter(struct osmo_fsm_inst *fi, uint32_t prev_
 		vlr_subscr_enable_expire_lu(vsub);
 	}
 
+	/* We no longer care about assignment responses. */
+	osmo_timer_del(&msc_a->cc.assignment_request_pending);
+
 	/* If we're closing in a middle of a trans, we need to clean up */
 	trans_conn_closed(msc_a);
 
@@ -997,6 +1055,7 @@ static void msc_a_fsm_released(struct osmo_fsm_inst *fi, uint32_t event, void *d
 void msc_a_fsm_cleanup(struct osmo_fsm_inst *fi, enum osmo_fsm_term_cause cause)
 {
 	struct msc_a *msc_a = msc_a_fi_priv(fi);
+	struct vlr_subscr *vsub = msc_a_vsub(msc_a);
 
 	trans_conn_closed(msc_a);
 
@@ -1004,6 +1063,13 @@ void msc_a_fsm_cleanup(struct osmo_fsm_inst *fi, enum osmo_fsm_term_cause cause)
 		LOG_MSC_A(msc_a, LOGL_ERROR, "Deallocating active transactions failed\n");
 
 	LOG_MSC_A_CAT(msc_a, DREF, LOGL_DEBUG, "max total use count was %d\n", msc_a->max_total_use_count);
+
+	/* Invalidate the active conn in VLR subscriber state, if any. */
+	if (vsub && vsub->msc_conn_ref == msc_a)
+		vsub->msc_conn_ref = NULL;
+
+	osmo_timer_del(&msc_a->lu_delay_timer);
+	osmo_timer_del(&msc_a->cc.assignment_request_pending);
 }
 
 const struct value_string msc_a_fsm_event_names[] = {
@@ -1218,6 +1284,7 @@ struct msc_a *msc_a_alloc(struct msub *msub, struct ran_infra *ran)
 	osmo_use_count_make_static_entries(&msc_a->use_count, msc_a->use_count_buf, ARRAY_SIZE(msc_a->use_count_buf));
 	/* Start timeout for first state */
 	msc_a_state_chg_always(msc_a, MSC_A_ST_VALIDATE_L3);
+	osmo_timer_setup(&msc_a->lu_delay_timer, &lu_delay_timer_cb, msc_a);
 	return msc_a;
 }
 
@@ -1431,6 +1498,9 @@ static void msc_a_up_call_assignment_complete(struct msc_a *msc_a, const struct
 	const struct gsm0808_speech_codec *codec_if_known = ac->assignment_complete.codec_present ?
 							    &ac->assignment_complete.codec : NULL;
 
+	/* Pending assignment has worked out. We're no longer waiting for a response now. */
+	osmo_timer_del(&msc_a->cc.assignment_request_pending);
+
 	/* For a voice group call, handling is performed by VGCS FSM */
 	gcc_trans = trans_find_by_type(msc_a, TRANS_GCC);
 	if (gcc_trans) {
@@ -1460,7 +1530,23 @@ static void msc_a_up_call_assignment_complete(struct msc_a *msc_a, const struct
 	}
 
 	if (codec_if_known) {
-		const struct codec_mapping *codec_cn;
+		const struct codec_mapping *codec_assigned;
+
+		/* Check for unexpected codec with CSD */
+		switch (cc_trans->bearer_cap.transfer) {
+		case GSM48_BCAP_ITCAP_FAX_G3:
+		case GSM48_BCAP_ITCAP_3k1_AUDIO:
+		case GSM48_BCAP_ITCAP_UNR_DIG_INF:
+			if (codec_if_known->type == GSM0808_SCT_CSD)
+				break; /* we're good */
+			LOG_TRANS(cc_trans, LOGL_ERROR, "Unexpected codec in Assignment Complete for CSD: %s\n",
+				  gsm0808_speech_codec_type_name(codec_if_known->type));
+			call_leg_release(msc_a->cc.call_leg);
+			return;
+		default:
+			break;
+		}
+
 		/* For 2G:
 		 * - The Assignment Complete has returned a specific codec (e.g. FR3 for AMR FR).
 		 * - Set this codec at the MGW endpoint facing the RAN.
@@ -1474,31 +1560,22 @@ static void msc_a_up_call_assignment_complete(struct msc_a *msc_a, const struct
 		 * - ran_msg_iu.c always returns FR3 (AMR FR) for the assigned codec. Set that at the MGW towards CN.
 		 * - So the MGW decapsulates IuUP <-> AMR
 		 */
-		codec_cn = codec_mapping_by_gsm0808_speech_codec_type(codec_if_known->type);
+		codec_assigned = codec_mapping_by_gsm0808_speech_codec_type(codec_if_known->type);
 		/* TODO: use codec_mapping_by_gsm0808_speech_codec() to also match on codec_if_known->cfg */
-		if (!codec_cn) {
+		if (!codec_assigned) {
 			LOG_TRANS(cc_trans, LOGL_ERROR, "Unknown codec in Assignment Complete: %s\n",
 				  gsm0808_speech_codec_type_name(codec_if_known->type));
 			call_leg_release(msc_a->cc.call_leg);
 			return;
 		}
 
-		/* Check for unexpected codec with CSD */
-		if (cc_trans->bearer_cap.transfer == GSM48_BCAP_ITCAP_UNR_DIG_INF &&
-		    codec_if_known->type != GSM0808_SCT_CSD) {
-			LOG_TRANS(cc_trans, LOGL_ERROR, "Unexpected codec in Assignment Complete for CSD: %s\n",
-				  gsm0808_speech_codec_type_name(codec_if_known->type));
-			call_leg_release(msc_a->cc.call_leg);
-			return;
-		}
-
 		/* Update RAN-side endpoint CI from Assignment result -- unless it is forced by the ran_infra, in which
 		 * case it remains unchanged as passed to the earlier call of call_leg_ensure_ci(). */
 		if (msc_a->c.ran->force_mgw_codecs_to_ran.count == 0)
-			rtp_stream_set_one_codec(rtps_to_ran, &codec_cn->sdp);
+			rtp_stream_set_one_codec(rtps_to_ran, &codec_assigned->sdp);
 
 		/* Update codec filter with Assignment result, for the CN side */
-		cc_trans->cc.codecs.assignment = codec_cn->sdp;
+		cc_trans->cc.codecs.assignment = codec_assigned->sdp;
 	} else {
 		/* No codec passed in Assignment Complete, set 'codecs.assignment' to none. */
 		cc_trans->cc.codecs.assignment = (struct sdp_audio_codec){};
@@ -1527,10 +1604,15 @@ static void msc_a_up_call_assignment_complete(struct msc_a *msc_a, const struct
 	}
 }
 
+/* Invoked when Assignment has failed, either by a failure response, or by timeout. When failing on timeout,
+ * pass af == NULL. */
 static void msc_a_up_call_assignment_failure(struct msc_a *msc_a, const struct ran_msg *af)
 {
 	struct gsm_trans *trans;
 
+	/* Pending assignment has failed. We're no longer waiting for a response now. */
+	osmo_timer_del(&msc_a->cc.assignment_request_pending);
+
 	/* For a normal voice call, there will be an rtp_stream FSM. */
 	if (msc_a->cc.call_leg && msc_a->cc.call_leg->rtp[RTP_TO_RAN]) {
 		LOG_MSC_A(msc_a, LOGL_ERROR, "Assignment Failure, releasing call\n");
@@ -1562,6 +1644,12 @@ static void msc_a_up_call_assignment_failure(struct msc_a *msc_a, const struct r
 	msc_a_release_cn(msc_a);
 }
 
+static void assignment_request_timeout_cb(void *data)
+{
+	struct msc_a *msc_a = data;
+	msc_a_up_call_assignment_failure(msc_a, NULL);
+}
+
 static void msc_a_up_classmark_update(struct msc_a *msc_a, const struct osmo_gsm48_classmark *classmark,
 				      struct osmo_gsm48_classmark *dst)
 {
